/*
 * multi-app-packet-sink.cc
 *
 *  Created on: Oct 15, 2015
 *    Author: Ali Rostami (WINLAB, Rutgers University)
 *    Email: rostami@winlab.rutgers.edu
 *    Homepage: alirostami.info
 */

#define _USE_MATH_DEFINES
#include <math.h>
#include <sys/stat.h>

#include <iostream>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include "ns3/address.h"
#include "ns3/address-utils.h"
#include "ns3/log.h"
#include "ns3/inet-socket-address.h"
#include "ns3/node.h"
#include "ns3/socket.h"
#include "ns3/udp-socket.h"
#include "ns3/packet-socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/udp-socket-factory.h"
#include "multi-app-packet-sink.h"
#include "cbp.h"
#include "multi-app.h"
#include <ns3/multiapp-packet.h>
#include "ns3/node-list.h"

using namespace std;

namespace ns3 {

NS_LOG_COMPONENT_DEFINE ("MultiAppPacketSink");
NS_OBJECT_ENSURE_REGISTERED (MultiAppPacketSink);

TypeId
MultiAppPacketSink::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::MultiAppPacketSink")
    .SetParent<Application> ()
    .AddConstructor<MultiAppPacketSink> ()
    .AddAttribute ("Local", "The Address on which to Bind the rx socket.",
                   AddressValue (),
                   MakeAddressAccessor (&MultiAppPacketSink::m_local),
                   MakeAddressChecker ())
    .AddAttribute ("Protocol", "The type id of the protocol to use for the rx socket.",
                   TypeIdValue (UdpSocketFactory::GetTypeId ()),
                   MakeTypeIdAccessor (&MultiAppPacketSink::m_tid),
                   MakeTypeIdChecker ())
  ;
  return tid;
}

MultiAppPacketSink::MultiAppPacketSink ()
{
  NS_LOG_FUNCTION (this);
  m_socket = 0;
  m_totalRx = 0;
  rssi_ = 0.0;
  for(int i=0;i<5000;i++)
    {
      sight_case[i] = -1;
    }
}

MultiAppPacketSink::~MultiAppPacketSink()
{
  NS_LOG_FUNCTION (this);

}

Ptr<Socket>
MultiAppPacketSink::GetListeningSocket (void) const
{
  NS_LOG_FUNCTION (this);
  return m_socket;
}

std::list<Ptr<Socket> >
MultiAppPacketSink::GetAcceptedSockets (void) const
{
  NS_LOG_FUNCTION (this);
  return m_socketList;
}

void MultiAppPacketSink::DoDispose (void)
{
  NS_LOG_FUNCTION (this);
  m_socket = 0;
  m_socketList.clear ();

  // chain up
  Application::DoDispose ();
}


// Application Methods
void MultiAppPacketSink::StartApplication ()    // Called at time specified by Start
{
  NS_LOG_FUNCTION (this);
  // Create the socket if not already
  if (!m_socket)
    {
      m_socket = Socket::CreateSocket (GetNode (), m_tid);
      m_socket->Bind (m_local);
      m_socket->Listen ();
      m_socket->ShutdownSend ();
      if (addressUtils::IsMulticast (m_local))
        {
          Ptr<UdpSocket> udpSocket = DynamicCast<UdpSocket> (m_socket);
          if (udpSocket)
            {
              // equivalent to setsockopt (MCAST_JOIN_GROUP)
              udpSocket->MulticastJoinGroup (0, m_local);
            }
          else
            {
              NS_FATAL_ERROR ("Error: joining multicast on a non-UDP socket");
            }
        }
    }

  m_socket->SetRecvCallback (MakeCallback (&MultiAppPacketSink::HandleRead, this));
  m_socket->SetAcceptCallback (
    MakeNullCallback<bool, Ptr<Socket>, const Address &> (),
    MakeCallback (&MultiAppPacketSink::HandleAccept, this));
  m_socket->SetCloseCallbacks (
    MakeCallback (&MultiAppPacketSink::HandlePeerClose, this),
    MakeCallback (&MultiAppPacketSink::HandlePeerError, this));


  //get my Id
  node = m_socket->GetNode();
  nodeID = node->GetId();
  mobilityModel = node->GetObject<MobilityModel>();
}

void MultiAppPacketSink::StopApplication ()     // Called at time specified by Stop
{
  NS_LOG_FUNCTION (this);

  fclose(trcFile);
  while(!m_socketList.empty ()) //these are accepted sockets, close them
    {
      Ptr<Socket> acceptedSocket = m_socketList.front ();
      m_socketList.pop_front ();
      acceptedSocket->Close ();
    }
  if (m_socket)
    {
      m_socket->Close ();
      m_socket->SetRecvCallback (MakeNullCallback<void, Ptr<Socket> > ());
    }

   Simulator::Cancel (RXMEvent);
}

void MultiAppPacketSink::HandleRead (Ptr<Socket> socket)
{

  NS_LOG_FUNCTION (this << socket);
  Ptr<Packet> packet;
  Address from;
  while (packet = socket->RecvFrom (from))
    {
      if (packet->GetSize () == 0)
        { //EOF
          break;
        }
      if (InetSocketAddress::IsMatchingType (from))
        {
          m_totalRx += packet->GetSize ();
          InetSocketAddress address = InetSocketAddress::ConvertFrom (from);
          (void) address;
        }
      receiveLog(packet, socket);
      m_rxTrace (packet, from);
    }
}

void
MultiAppPacketSink::receiveLog(Ptr<Packet> packet,Ptr<Socket> socket)
{
  double now = Simulator::Now ().GetSeconds();

  PayloadMultiApp payload;
  packet->CopyData((uint8_t *)&payload,sizeof(PayloadMultiApp));
  Ptr<Node> sender = NodeList::GetNode(GetNodeID());

  Ptr<MultiApp> sender_app = DynamicCast<MultiApp>(sender->GetApplication(1));

  double current_X = (double)(mobilityModel->GetPosition().x);
  double current_Y = (double)(mobilityModel->GetPosition().y);
  if(sender_app->GetNodeType()!=1)
    {
      std::cout<<"r "<<now<<" "
    		   << nodeID <<" "
			   << current_X <<" "
			   << current_Y <<" "
			   << payload.node_ID <<" "
			   << packet->GetUid()
			   << " rssi " << rssi_ <<" "
			   << sender_app->GetNodeType()<< " "
			   << sight_case[payload.node_ID]
			   << std::endl;
    }
  sight_case[payload.node_ID] = -1;
}

void
MultiAppPacketSink::HandlePeerClose (Ptr<Socket> socket)
{
  NS_LOG_INFO ("PktSink, peerClose");
}

void
MultiAppPacketSink::HandlePeerError (Ptr<Socket> socket)
{
  NS_LOG_INFO ("PktSink, peerError");
}


void
MultiAppPacketSink::HandleAccept (Ptr<Socket> s, const Address& from)
{
  NS_LOG_FUNCTION (this << s << from);
  s->SetRecvCallback (MakeCallback (&MultiAppPacketSink::HandleRead, this));
  m_socketList.push_back (s);
}

void
RxDrop (Ptr<Node>node, Ptr<const Packet> p)
{

  Ptr<MultiAppPacketSink> sink_app = DynamicCast<MultiAppPacketSink>(node->GetApplication(0));
  Ptr<MultiApp> gen_app = DynamicCast<MultiApp>(node->GetApplication(1));

  double now = Simulator::Now ().GetSeconds();

  WifiMacHeader header;
  p->PeekHeader(header);
  Mac48Address addr = header.GetAddr2();
  uint32_t sourceId = ConvertMacAddressToId (addr); // XXX: MAC Address is starting from 1 - N (N is number of nodes)

  Ptr<MobilityModel> mob = node->GetObject<MobilityModel>();

  Ptr<Node> source_node = NodeList::GetNode(sourceId-1);
  Ptr<MobilityModel> source_mob = source_node->GetObject<MobilityModel>();

  double current_X = ((mob->GetPosition()).x);
  double current_Y = ((mob->GetPosition()).y);

  double source_X = ((source_mob->GetPosition()).x);
  double source_Y = ((source_mob->GetPosition()).y);

  double dis = std::sqrt(std::pow(source_X - current_X,2) + std::pow(source_Y - current_Y,2));

  // We just want to keep logs for the communications where the transmitters are at most 500 m away
  if(dis < 500 && gen_app->GetNodeType()!=1)
  {
    std::cout<<"d "<<now<<" "<< sink_app->GetNodeID() <<" "<< current_X <<" "<< current_Y <<" "<< sourceId-1 <<" "<< p->GetUid() <<" "<< gen_app->GetNodeType()<< " " << sink_app->sight_case[sourceId-1] << std::endl;
  }

  sink_app->sight_case[sourceId-1] = -1;
}

// To get the Sender ID number of a Packet With MAC header
uint32_t ConvertMacAddressToId(Mac48Address macAddr)
{
    uint8_t m_address[6];
    macAddr.CopyTo(m_address);
    uint64_t id = 0;
    for (int i=0;i<6;i++) id=(id<<8)+m_address[i];
    return id;
}

void RssiTrace (Ptr<Node> node, double RSSI)
{
  Ptr<MultiAppPacketSink> sink= DynamicCast<MultiAppPacketSink>(node->GetApplication(0));
  sink->rssi_ = RSSI;
}

} // Namespace ns3


