/*
 * multi-app.cc
 *
 *  Created on: Oct 15, 2015
 *    Author: Ali Rostami (WINLAB, Rutgers University)
 *    Email: rostami@winlab.rutgers.edu
 *    Homepage: alirostami.info
 */

#include <cmath>

#include "multi-app.h"
#include "ns3/core-module.h"
#include <ns3/application.h>
#include <ns3/log.h>
#include <ns3/multiapp-packet.h>
#include "ns3/udp-socket-factory.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/random-variable.h"


NS_LOG_COMPONENT_DEFINE ("MultiApp");

namespace ns3{

//================================ MultiApp class ================================//

NS_OBJECT_ENSURE_REGISTERED (MultiApp);

TypeId
MultiApp::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::MultiApp")
    .SetParent<Application> ()
    .AddConstructor<MultiApp> ()
    .AddAttribute ("DataRate", "The data rate in on state.",
					 DataRateValue (DataRate ("500kb/s")),
					 MakeDataRateAccessor (&MultiApp::dataRate_),
					 MakeDataRateChecker ())
    .AddAttribute("PacketSize", "The packet size for this application",
		       UintegerValue (512),
		       MakeUintegerAccessor (&MultiApp::packetSize),
		       MakeUintegerChecker<uint32_t> ())
    .AddAttribute("CBPCalPeriod", "CBP log interval in millisecond",
		       UintegerValue (100),
		       MakeUintegerAccessor (&MultiApp::cbpCalPeriod),
		       MakeUintegerChecker<uint32_t> ())
    .AddAttribute("MessageRate", "Number of message that should be generate per second",
		       UintegerValue (10),
		       MakeUintegerAccessor (&MultiApp::msgRate_),
		       MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("Remote", "The address of the destination",
		       AddressValue (),
		       MakeAddressAccessor (&MultiApp::app_peer),
		       MakeAddressChecker ())
    .AddAttribute ("Protocol", "The type of protocol to use.",
		       TypeIdValue (UdpSocketFactory::GetTypeId ()),
		       MakeTypeIdAccessor (&MultiApp::m_tid),
		       MakeTypeIdChecker ())
    .AddAttribute("AppType", "Algorithm type",
		       UintegerValue (1),
		       MakeUintegerAccessor (&MultiApp::appType_),
		       MakeUintegerChecker<uint32_t> ())
    .AddAttribute("Totalnn", "Total Number of nodes in the simulation",
		       UintegerValue (0),
		       MakeUintegerAccessor (&MultiApp::nn_),
		       MakeUintegerChecker<uint32_t> ())
    .AddAttribute("TotalMonitorNodes", "Total Number of nodes in the simulation",
		       UintegerValue (0),
		       MakeUintegerAccessor (&MultiApp::number_of_monitor_nodes_),
		       MakeUintegerChecker<uint32_t> ())
    .AddAttribute("vehicleNum", "Total Number of nodes in the simulation",
		       UintegerValue (0),
		       MakeUintegerAccessor (&MultiApp::number_of_vehicles),
		       MakeUintegerChecker<uint32_t> ())
    .AddAttribute("SimType", "Total Number of nodes in the simulation",
		       UintegerValue (2),
		       MakeUintegerAccessor (&MultiApp::sim_type),
		       MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("sidewalkFile", "the path to the file consisting buildings corner coordination on the map",
		       StringValue ("MMMMMM"),
		       MakeStringAccessor(&MultiApp::sidewalkFilePath_),
		       MakeStringChecker())
    ;

  return tid;
}

MultiApp::MultiApp(void)
{
  NS_LOG_FUNCTION (this);
  app_maxBytes = 0;
  app_socket = 0;
  app_connected = false;
  qosTag.SetTid (uint32_t (2));
  msgSeq = 0;
  max_randomization = 1.0;

  jitter = CreateObject<UniformRandomVariable>();
  jitter->SetAttribute ("Min", DoubleValue (-0.005));
  jitter->SetAttribute ("Max", DoubleValue (0.005));

  randStart = CreateObject<UniformRandomVariable>();
  randStart->SetAttribute ("Min", DoubleValue (0));
  randStart->SetAttribute ("Max", DoubleValue (max_randomization));
}

MultiApp::~MultiApp()
{
  NS_LOG_FUNCTION (this);
}

void
MultiApp::SetMaxBytes (uint32_t maxBytes)
{
  NS_LOG_FUNCTION (this << maxBytes);
  app_maxBytes = maxBytes;
}

void
MultiApp::StartApplication()
{
  NS_LOG_FUNCTION (this);

  // Create the socket if not already
  if (!app_socket)
    {
      app_socket = Socket::CreateSocket (GetNode(), m_tid);
      app_socket->Bind ();
      app_socket->Connect (app_peer);
      app_socket->SetAllowBroadcast (true);
      app_socket->ShutdownRecv ();
    }

  node = GetNode ();
  nd = DynamicCast<WifiNetDevice>(node->GetDevice(0));
  mobilityModel = node->GetObject<MobilityModel>();
  nodeID_ = node -> GetId();
  qosTag.SetTid (uint32_t (2)); // background traffic category

  CBP_ = new CBP();
  CBP_->SetNode(node);
  CBP_->cbp_calculation_window_ = cbpCalPeriod;
  CBP_->CBPStart();

  if(nodeID_ < number_of_vehicles)
    {
      SetNodeType(0);   // This is a vehicle
    }
  else if(nodeID_ < nn_ - number_of_monitor_nodes_)
    {
      SetNodeType(1);   // This is a pedestrian
    }
  else
    {
      SetNodeType(2);   // This is a monitor node
    }

  if(appType_ == Baseline)
    {
      startBaseline();
    }
  else if(appType_ == OnStreetPed)
    {
      startStreetBaseline();
    }
  else if(appType_ == MovingPed)
    {
      startPedMoveBaseline();
    }
  else if(appType_ == MulTxRates)
    {
      startDifferentFreq();
    }
}

void
MultiApp::StopApplication()
{
  NS_LOG_FUNCTION (this);
  fclose(trcFile);
}

void
MultiApp::startBaseline()
{
  NS_LOG_FUNCTION (this);
//  nextCBPCalSchedule();
  firstSchedule();  
}

void
MultiApp::startStreetBaseline()
{
  NS_LOG_FUNCTION (this);
  sidewalk = new PolygonList();
  sidewalk->getAllPolygons(sidewalkFilePath_);
//  nextCBPCalSchedule();
  firstSchedule();
}

void
MultiApp::startPedMoveBaseline()
{
  NS_LOG_FUNCTION (this);
  max_idle = 2.0;
  last_x = (double)(mobilityModel->GetPosition().x);
  last_y = (double)(mobilityModel->GetPosition().y);
  last_movement = -max_idle;
  movement_threshold = 0.0;

  movementCheckTimer();
//  nextCBPCalSchedule();
  firstSchedule();
}

void
MultiApp::startDifferentFreq()
{
  NS_LOG_FUNCTION (this);

  msgRate_ = 5; //Hz
  msgRate2_ = 2; //Hz

  max_idle = 2.0;
  last_x = (double)(mobilityModel->GetPosition().x);
  last_y = (double)(mobilityModel->GetPosition().y);
  last_movement = -max_idle;
  movement_threshold = 0.0;

  movementCheckTimer();
//  nextCBPCalSchedule();
  firstSchedule();
}

void
MultiApp::firstSchedule()
{
  double st = randStart->GetValue();
  if(sim_type == 2 && nodeID_ < nn_ - number_of_monitor_nodes_)
    appStart = Simulator::Schedule(Seconds(st), &MultiApp::sendPacket, this);
  else if(sim_type == 1 && nodeID_ >= number_of_vehicles && nodeID_ < nn_ - number_of_monitor_nodes_)
    appStart = Simulator::Schedule(Seconds(st), &MultiApp::sendPacket, this);
  else if(sim_type == 0 && nodeID_ < number_of_vehicles)
    appStart = Simulator::Schedule(Seconds(st), &MultiApp::sendPacket, this);
}

void
MultiApp::startLIMERIC()
{

}

void
MultiApp::sendPacket ()
{
  NS_LOG_FUNCTION (this);

  double now = Simulator::Now ().GetSeconds();
  bool sent = false;

  PayloadMultiApp payload=PayloadMultiApp();
  Ptr<Packet> packet;

  payload = CreatePacket();


  if(appType_ == Baseline)
    {
      packet  = Create<Packet> ((uint8_t*) &payload,packetSize);
      packet->AddPacketTag(qosTag);
      app_socket->Send(packet);
      sent = true;
      nextTxSchedule();
    }

  else if(appType_ == OnStreetPed)
    {
      if(!(sidewalk->isOnSidewalk(payload.X,payload.Y)))
				{
					packet  = Create<Packet> ((uint8_t*) &payload,packetSize);
					packet->AddPacketTag(qosTag);
					app_socket->Send(packet);
					sent = true;
				}
      nextTxSchedule();
    }
  else if(appType_ == MovingPed)
    {
      if((node_type == 1 && last_movement >= now - max_idle) || node_type == 0)
				{
					packet  = Create<Packet> ((uint8_t*) &payload,packetSize);
					packet->AddPacketTag(qosTag);
					app_socket->Send(packet);
					sent = true;
				}
      nextTxSchedule();
    }
  else if(appType_ == MulTxRates)
    {
      if((node_type == 1 && last_movement >= now - max_idle) || node_type == 0)
				{
					packet  = Create<Packet> ((uint8_t*) &payload,packetSize);
					packet->AddPacketTag(qosTag);
					app_socket->Send(packet);
					sent = true;
					nextTxEvent = Simulator::Schedule(Seconds(1/(double)msgRate_+jitter->GetValue()), &MultiApp::sendPacket, this);
				}
      else if(node_type == 1 && last_movement < now - max_idle)
				{
					packet  = Create<Packet> ((uint8_t*) &payload,packetSize);
					packet->AddPacketTag(qosTag);
					app_socket->Send(packet);
					sent = true;
					nextTxEvent = Simulator::Schedule(Seconds(1/(double)msgRate2_+jitter->GetValue()), &MultiApp::sendPacket, this);
				}
    }

  if(sent)//log
    {
      std::cout<<"s "<<now<<" "
      		     << nodeID_ <<" "
							 << payload.X <<" "
							 << payload.Y <<" "
							 << payload.node_ID <<" "
							 << packet->GetUid() <<" "
							 << node_type
							 << std::endl ;
      msgSeq++; //packet sequence number increment
    }
}

void
MultiApp::nextTxSchedule()
{
  NS_LOG_FUNCTION (this);
  nextTxEvent = Simulator::Schedule(Seconds(1/(double)msgRate_+jitter->GetValue()), &MultiApp::sendPacket, this);
}

PayloadMultiApp MultiApp::CreatePacket()
{
  NS_LOG_FUNCTION (this);

  PayloadMultiApp payload = PayloadMultiApp();
  payload.node_ID = nodeID_;

  double now = Simulator::Now().GetSeconds ();
  payload.time = now;

	payload.X = (double)(mobilityModel->GetPosition().x);
	payload.Y = (double)(mobilityModel->GetPosition().y);
	payload.speed_x = (double)(mobilityModel->GetVelocity().x);
	payload.speed_y = (double)(mobilityModel->GetVelocity().y);
	payload.heading = atan2(payload.speed_y, payload.speed_x) * 180/M_PI;
	payload.seqNum= msgSeq;

  return payload;
}

void
MultiApp::movementCheckTimer()
{
  NS_LOG_FUNCTION (this);
  next_check = Simulator::Schedule(Seconds(1/(double)msgRate_), &MultiApp::movementChecker, this);
}

void
MultiApp::movementChecker()
{
  NS_LOG_FUNCTION (this);

  double my_x, my_y;
  my_x = (double)(mobilityModel->GetPosition().x);
  my_y = (double)(mobilityModel->GetPosition().y);

  if (sqrt(pow(my_x-last_x,2)+pow(my_y-last_y,2))>movement_threshold)
    {
      last_movement = Simulator::Now().GetSeconds ();
      last_x = my_x;
      last_y = my_y;
    }

  movementCheckTimer();
}

// Callback function to keep track of the idle time
void idleStateTraceCBP (Ptr<Node>node, Time start, Time duration, enum WifiPhy::State state)
{
  Ptr<MultiApp> app= DynamicCast<MultiApp>(node->GetApplication(1));

	double du;
	if(state == WifiPhy::IDLE)
		{
			if((app->CBP_)->start_time_ > start)
				{
					du = duration.GetSeconds()*1000 - ((app->CBP_)->start_time_ - start).GetSeconds()*1000;
				}
			else
				{
					du = duration.GetSeconds()*1000;
				}
			(app->CBP_)->idle_duration_ += du;
		}
	else
		{
			std::cerr<<"idleStateTrace is not correctly triggered!\n"<<std::endl;
			exit(1);
		}
}

} // ns3 namespace
